"""
Clunky pepperings-over of communication with containers.

Some of this is warts from aiodocker, but mixed in with fun
special-to-this-package logic occasionally.
"""

from __future__ import annotations

from collections import deque
from contextlib import asynccontextmanager, suppress
from typing import TYPE_CHECKING
import asyncio
import json

from aiodocker import Docker
from attrs import field, frozen, mutable
import aiodocker.exceptions

from bowtie._core import (
    GotStderr,
    InvalidResponse,
    NoSuchImplementation,
    Restarted,
    StartupFailed,
)

if TYPE_CHECKING:
    from collections.abc import AsyncIterator, Awaitable
    from contextlib import AbstractAsyncContextManager
    from typing import Any, Self

    import aiodocker.containers
    import aiodocker.docker
    import aiodocker.stream  # noqa: TCH004 ??? no it's not?

    from bowtie._commands import Message


IMAGE_REPOSITORY = "ghcr.io/bowtie-json-schema"


@frozen
class _ClosedStream(Exception):
    """
    The stream is closed, and we tried to send something on it.

    This exception should never bubble out of this module, it should be handled
    by the connection.
    """

    container: aiodocker.containers.DockerContainer


@mutable
class Stream:
    """
    Wrapper to make aiodocker's Stream more pleasant to use.
    """

    _stream: aiodocker.stream.Stream = field(repr=False, alias="stream")
    _container: aiodocker.containers.DockerContainer = field(
        repr=False,
        alias="container",
    )
    _read_timeout_sec: float | None = field(
        repr=False,
        alias="read_timeout_sec",
    )
    _buffer: deque[bytes] = field(factory=deque, alias="buffer")
    _last: bytes = b""

    @classmethod
    def attached_to(
        cls,
        container: aiodocker.containers.DockerContainer,
        **kwargs: Any,
    ) -> Stream:
        stream = container.attach(stdin=True, stdout=True, stderr=True)
        return cls(stream=stream, container=container, **kwargs)

    def _read_with_timeout(self) -> Awaitable[aiodocker.stream.Message | None]:
        read = self._stream.read_out()
        return asyncio.wait_for(read, timeout=self._read_timeout_sec)

    async def send(self, message: str) -> None:
        try:  # aiodocker doesn't appear to properly report stream closure
            await self._stream.write_in(message.encode())
        except aiodocker.exceptions.DockerError as err:
            raise _ClosedStream(self._container) from err
        except AttributeError:
            raise _ClosedStream(self._container) from None

    async def receive(self) -> str:
        if self._buffer:
            return self._buffer.popleft().decode()

        while True:
            try:
                message = await self._read_with_timeout()
            except asyncio.exceptions.TimeoutError:
                info: dict[str, Any] = await self._container.show()  # type: ignore[reportUnknownMemberType]
                if not info["State"]["Running"]:
                    raise _ClosedStream(self._container)
                raise

            if message is not None:
                break
            info: dict[str, Any] = await self._container.show()  # type: ignore[reportUnknownMemberType]
            if not info["State"]["Running"]:
                raise _ClosedStream(self._container)

        if message.stream == 2:  # type: ignore[reportUnknownMemberType]  # noqa: PLR2004
            data: list[bytes] = []

            while message.stream == 2:  # type: ignore[reportUnknownMemberType]  # noqa: PLR2004
                data.append(message.data)  # type: ignore[reportUnknownMemberType]
                try:
                    message = await self._read_with_timeout()
                except asyncio.exceptions.TimeoutError:
                    message = None
                if message is None:
                    raise GotStderr(b"".join(data))

        line: bytes
        rest: list[bytes]
        while True:
            line, *rest = message.data.split(b"\n")  # type: ignore[reportUnknownMemberType]
            if rest:
                line, self._last = self._last + line, rest.pop()  # type: ignore[reportUnknownVariableType]
                self._buffer.extend(rest)
                return line.decode()  # type: ignore[reportUnknownVariableType]

            message = None
            while message is None:
                message = await self._read_with_timeout()
            self._last += line  # type: ignore[reportUnknownMemberType]

    async def ensure_deleted(self):
        with suppress(aiodocker.exceptions.DockerError):
            await self._container.delete(force=True)  # type: ignore[reportUnknownMemberType]


@mutable
class Connection:
    """
    A connection with a restartable container over stdio via request/responses.

    Requests and responses are JSON-serializable messages, with serialization
    handled here.
    """

    _image: str = field(alias="image")

    _docker: aiodocker.docker.Docker = field(repr=False, alias="docker")
    _stream: Stream = field(default=None, repr=False, alias="stream")

    # Maybe second versions of these will be useful also at the Implementation
    # level again, to control for non-protocol-related flakiness or slowness
    _restarts: int = field(default=10, repr=False, alias="restarts")
    _read_timeout_sec: float | None = field(
        default=2.0,
        converter=lambda value: value or None,  # type: ignore[reportUnknownLambdaType]
        repr=False,
    )

    #: A per-request number of retries, before giving up
    _retry: int = field(default=3, repr=False)

    @classmethod
    @asynccontextmanager
    async def open(
        cls,
        image_name: str,
        **kwargs: Any,
    ) -> AsyncIterator[Self]:
        async with Docker() as docker:
            self = cls(docker=docker, image=image_name, **kwargs)

            try:
                await self._start_container_maybe_pull()
            except GotStderr as error:
                err = StartupFailed(
                    name=image_name,
                    stderr=error.stderr.decode(),
                )
                raise err from None
            except _ClosedStream:
                raise StartupFailed(name=image_name) from None

            yield self
            await self._stream.ensure_deleted()

    async def _start_container_maybe_pull(self):
        # You would think we would use aiodocker's container.start() function
        # which essentially does the below. You would think wrong.
        # That function will pull the *entire* image repository if it ends up
        # pulling our harness image -- so here we reimplement it, but only
        # pull :latest when the image is missing.
        try:
            await self._start_container()
        except aiodocker.exceptions.DockerError as err:
            if err.status != 404:  # noqa: PLR2004
                raise
            try:
                await self._docker.pull(from_image=self._image, tag="latest")  # type: ignore[reportUnknownMemberType]
            except aiodocker.exceptions.DockerError as err:
                # This craziness can go wrong in various ways, none of them
                # machine parseable.

                status, data, *_ = err.args
                if data.get("cause") == "image not known":
                    raise NoSuchImplementation(self._image) from err

                message = ghcr = data.get("message", "")

                if status == 500:  # noqa: PLR2004
                    try:
                        # GitHub Registry saying an image doesn't exist as
                        # reported within GitHub Actions' version of Podman...
                        # This is some crazy string like:
                        #   Head "https://ghcr.io/v2/bowtie-json-schema/image-name/manifests/latest": denied  # noqa: E501
                        # with seemingly no other indication elsewhere and
                        # obviously no good way to detect this specific case
                        no_image = message.endswith('/latest": denied')
                    except Exception:  # noqa: BLE001, S110
                        pass
                    else:
                        if no_image:
                            raise NoSuchImplementation(self._image)

                    try:
                        # GitHub Registry saying an image doesn't exist as
                        # reported locally via podman on macOS...

                        # message will be ... a JSON string !?! ...
                        error = json.loads(ghcr).get("message", "")
                    except Exception:  # noqa: BLE001, S110
                        pass  # nonJSON / missing key
                    else:
                        if "403 (forbidden)" in error.casefold():
                            raise NoSuchImplementation(self._image)

                raise StartupFailed(name=self._image, data=data) from err
            await self._start_container()

    async def _start_container(self):
        config = dict(
            Image=self._image,
            OpenStdin=True,
            HostConfig=dict(NetworkMode="none"),
        )
        # FIXME: name + labels
        container = await self._docker.containers.create(config=config)  # type: ignore[reportUnknownMemberType]
        await container.start()  # type: ignore[reportUnknownMemberType]
        self._stream = Stream.attached_to(
            container,
            read_timeout_sec=self._read_timeout_sec,
        )

    async def request(self, message: Message) -> Message | None:
        request = f"{json.dumps(message)}\n"

        try:
            await self._stream.send(request)
        except _ClosedStream:
            self._restarts -= 1
            await self._stream.ensure_deleted()
            await self._start_container()
            raise Restarted() from None

        for _ in range(self._retry):
            try:
                response = await self._stream.receive()
            except asyncio.exceptions.TimeoutError:
                continue
            except _ClosedStream as err:
                stderr: list[str] = await err.container.log(stderr=True)  # type: ignore[reportUnknownVariableType]
                if stderr:
                    raise GotStderr("".join(stderr).encode())  # type: ignore[reportUnknownVariableType]
                return

            try:
                return json.loads(response)
            except json.JSONDecodeError as err:
                raise InvalidResponse(contents=response) from err

    async def poison(self, message: dict[str, Any]) -> None:
        request = f"{json.dumps(message)}\n"
        with suppress(_ClosedStream):
            await self._stream.send(request)


@frozen
class ConnectableImage:

    _id: str = field(
        converter=lambda value: (  # type: ignore[reportUnknownLambdaType]
            value if "/" in value else f"{IMAGE_REPOSITORY}/{value}"
        ),
        alias="id",
    )

    connector = "image"

    def connect(self) -> AbstractAsyncContextManager[Connection]:
        return Connection.open(image_name=self._id)



from __future__ import annotations

from datetime import datetime, timezone
from typing import TYPE_CHECKING, TypedDict
import importlib.metadata
import json
import sys

from attrs import asdict, field, frozen
from attrs.filters import exclude
from rpds import HashTrieMap
import structlog.stdlib

from bowtie._commands import (
    Seq,
    SeqCase,
    SeqResult,
    StartedDialect,
    Unsuccessful,
)
from bowtie._core import Dialect, TestCase, validator_registry

if TYPE_CHECKING:
    from collections.abc import Callable, Iterable, Mapping, Sequence
    from typing import Any, Literal, Self, TextIO

    from bowtie._commands import AnyTestResult, ImplementationId
    from bowtie._core import Example, ImplementationInfo, Test


class InvalidReport(Exception):
    """
    The report is invalid.
    """


class EmptyReport(InvalidReport):
    """
    The report was totally empty.
    """


class DuplicateCase(InvalidReport):
    """
    A `Seq` appeared twice in the report.
    """


class MissingFooter(InvalidReport):
    """
    A report is missing its footer.

    Even though that only tells us whether the report failed fast, it might
    mean there's actual data missing too.
    """


def writer(file: TextIO = sys.stdout) -> Callable[..., Any]:
    return lambda **result: file.write(f"{json.dumps(result)}\n")  # type: ignore[reportUnknownArgumentType]


@frozen
class Reporter:
    _write: Callable[..., Any] = field(default=writer(), alias="write")
    _log: structlog.stdlib.BoundLogger = field(
        factory=structlog.stdlib.get_logger,
    )

    def unacknowledged_dialect(
        self,
        implementation: str,
        dialect: Dialect,
        response: StartedDialect,
    ):
        self._log.warn(
            (
                "Implicit dialect not acknowledged. "
                "Proceeding, but implementation may not have configured "
                "itself to handle schemas without $schema."
            ),
            logger_name=implementation,
            dialect=dialect.pretty_name,
            response=response,
        )

    def ready(self, run_metadata: RunMetadata):
        self._log.debug("Will speak", dialect=run_metadata.dialect)
        self._write(**run_metadata.serializable())

    def finished(self, did_fail_fast: bool):
        self._write(did_fail_fast=did_fail_fast)

    def case_started(self, seq_case: SeqCase, dialect: Dialect):
        self._write(**seq_case.serializable())
        log = self._log.bind(
            case=seq_case.case.description,
            schema=seq_case.case.schema,
        )

        if not seq_case.matches_dialect(dialect):
            log.warning(
                "$schema keyword does not seem to match the expected dialect",
                expected=dialect,
            )

        def got_result(result: SeqResult):
            bound = log.bind(logger_name=result.implementation)
            serialized = result.log_and_be_serialized(log=bound)
            self._write(**serialized)

        return got_result


@frozen
class RunMetadata:
    dialect: Dialect
    implementations: Sequence[ImplementationInfo] = field(
        repr=lambda value: (
            f"({len(value)} implementation{'s' if len(value) != 1 else ''})"
        ),
        alias="implementations",
    )
    bowtie_version: str = field(
        default=importlib.metadata.version("bowtie-json-schema"),
        eq=False,
        repr=False,
    )
    metadata: Mapping[str, Any] = field(factory=dict, repr=False)
    started: datetime = field(
        factory=lambda: datetime.now(timezone.utc),
        eq=False,
        repr=False,
    )

    @classmethod
    def from_dict(
        cls,
        dialect: str,
        implementations: dict[str, dict[str, Any]],
        started: str | None = None,
        **kwargs: Any,
    ) -> RunMetadata:
        from bowtie._core import ImplementationInfo

        if started is not None:
            kwargs["started"] = datetime.fromisoformat(started)
        return cls(
            dialect=Dialect.from_str(dialect),
            implementations=[
                ImplementationInfo.from_dict(image=image, **data)
                for image, data in implementations.items()
            ],
            **kwargs,
        )

    def serializable(self):
        as_dict = asdict(
            self,
            filter=exclude("dialect"),
            recurse=False,
        )
        as_dict.update(
            dialect=self.dialect.serializable(),
            started=as_dict.pop("started").isoformat(),
            # FIXME: This transformation is to support the UI parsing
            implementations={
                i.id: i.serializable() for i in self.implementations
            },
        )
        return as_dict


@frozen(eq=False)
class Report:
    r"""
    A Bowtie report, containing (amongst other things) results about cases run.

    In general, reports are assumed to be constructed by calling
    `Report.from_serialized` or an equivalent (i.e. by 'replaying' JSON output
    that was produced by Bowtie).

    When comparing reports (e.g. for equality), only the relative order of
    test cases is considered, not the exact `Seq`\ s used in the run.
    """

    _cases: HashTrieMap[Seq, TestCase] = field(alias="cases", repr=False)
    _results: HashTrieMap[
        ImplementationId,
        HashTrieMap[Seq, SeqResult],
    ] = field(
        repr=False,
        alias="results",
    )
    metadata: RunMetadata
    did_fail_fast: bool

    def __eq__(self, other: object):
        if type(other) is not Report:
            return NotImplemented

        if (  # short circuit for fewer/more cases or differing implementations
            len(self._cases) != len(other._cases)
            or self._results.keys() != other._results.keys()
        ):
            return False

        this, that = asdict(self, recurse=False), asdict(other, recurse=False)

        cases = [v for _, v in sorted(this.pop("_cases").items())]
        if cases != [v for _, v in sorted(that.pop("_cases").items())]:
            return False

        other_results = that.pop("_results")
        for name, results in this.pop("_results").items():
            ordered = [v for _, v in sorted(results.items())]
            if ordered != [v for _, v in sorted(other_results[name].items())]:
                return False
        return this == that

    @classmethod
    def from_input(cls, input: Iterable[Mapping[str, Any]]) -> Self:
        # TODO: Support some interface for enabling/disabling validation.
        validator = validator_registry().for_uri(
            "tag:bowtie.report,2024:report",
        )

        iterator = iter(input)
        header = next(iterator, None)
        if header is None:
            raise EmptyReport()
        validator.validate(header)
        metadata = RunMetadata.from_dict(**header)

        validator = metadata.dialect.current_dialect_resource() @ validator

        results: HashTrieMap[
            ImplementationId,
            HashTrieMap[Seq, SeqResult],
        ] = HashTrieMap.fromkeys(  # type: ignore[reportUnknownMemberType]
            (each.id for each in metadata.implementations),
            HashTrieMap(),
        )
        cases: HashTrieMap[Seq, TestCase] = HashTrieMap()

        for data in iterator:
            validator.validate(data)
            match data:
                case {"seq": seq, "case": case}:
                    if seq in cases:
                        raise DuplicateCase(seq)
                    case = TestCase.from_dict(dialect=metadata.dialect, **case)
                    cases = cases.insert(seq, case)
                    continue
                case {"did_fail_fast": did_fail_fast}:
                    return cls(
                        results=results,
                        cases=cases,
                        metadata=metadata,
                        did_fail_fast=did_fail_fast,
                    )
                case _:
                    result = SeqResult.from_dict(**data)

            current = results.get(result.implementation, HashTrieMap())
            results = results.insert(  # TODO: Complain if present
                result.implementation,
                current.insert(result.seq, result),
            )

        raise MissingFooter()

    @classmethod
    def from_serialized(cls, serialized: Iterable[str]) -> Self:
        return cls.from_input(json.loads(line) for line in serialized)

    @classmethod
    def empty(
        cls,
        implementations: Sequence[ImplementationInfo] = (),
        **kwargs: Any,
    ):
        """
        'The' empty report.
        """
        return cls(
            cases=HashTrieMap(),
            results=HashTrieMap(),
            metadata=RunMetadata(implementations=implementations, **kwargs),
            did_fail_fast=False,
        )

    @property
    def implementations(self) -> Sequence[ImplementationInfo]:
        return self.metadata.implementations

    @property
    def is_empty(self):
        return not self._cases

    @property
    def total_tests(self):
        return sum(len(case.tests) for case in self._cases.values())

    def unsuccessful(self, implementation: ImplementationId) -> Unsuccessful:
        """
        A count of the unsuccessful tests for the given implementation.
        """
        results = self._results[implementation].values()
        return sum((each.unsuccessful() for each in results), Unsuccessful())

    def worst_to_best(self):
        """
        All implementations ordered by number of unsuccessful tests.

        Ties are then broken alphabetically.
        """
        unsuccessful = [
            (implementation, self.unsuccessful(implementation.id))
            for implementation in self.implementations
        ]
        unsuccessful.sort(key=lambda each: (each[1].total, each[0].name))
        return unsuccessful

    def cases_with_results(
        self,
    ) -> Iterable[
        tuple[
            TestCase,
            Iterable[tuple[Example | Test, Mapping[str, AnyTestResult]]],
        ]
    ]:
        for seq, case in sorted(self._cases.items()):
            test_results: list[
                tuple[Example | Test, Mapping[str, AnyTestResult]]
            ] = []
            for i, test in enumerate(case.tests):
                test_result = {
                    each.id: self._results[each.id][seq].result_for(i)
                    for each in self.implementations
                }
                test_results.append((test, test_result))
            yield case, test_results

    def compliance_badges(self) -> Iterable[tuple[ImplementationInfo, Badge]]:
        for implementation in self.implementations:
            unsuccessful = self.unsuccessful(implementation.id)
            passed = self.total_tests - unsuccessful.total
            percentage = 100 * (passed / self.total_tests)
            r, g, b = 100 - int(percentage), int(percentage), 0
            yield implementation, Badge(
                schemaVersion=1,
                label=self.metadata.dialect.pretty_name,
                message="%d%% Passing" % int(percentage),
                color=f"{r:02x}{g:02x}{b:02x}",
            )


class Badge(TypedDict):
    schemaVersion: Literal[1]
    label: str
    message: str
    color: str


def supported_version_badge(dialects: Iterable[Dialect]) -> Badge:
    message = ", ".join(
        each.pretty_name.removeprefix("Draft ")
        for each in sorted(dialects, reverse=True)
    )
    return Badge(
        schemaVersion=1,
        label="JSON Schema Versions",
        message=message,
        color="lightgreen",
    )
